{-# LANGUAGE NumericUnderscores #-}
{-|
Module      : PUBMED
Description : PubMed API integration
Copyright   : (c) CNRS, 2019
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX
-}


module PUBMED where

import Conduit
import Control.Applicative
import Control.Concurrent (threadDelay)
import Control.Monad.Reader (ask, runReaderT)
import Data.Attoparsec.ByteString
import Data.Attoparsec.ByteString.Char8 (anyChar)
import Data.ByteString.Char8 (pack)
import Data.ByteString.Lazy (ByteString)
import Data.Maybe (fromMaybe)
import Data.Text (Text)
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import PUBMED.Client
import PUBMED.Parser
import Prelude hiding (takeWhile)
import Servant.Client (runClientM, mkClientEnv, BaseUrl(..), ClientEnv(..), ClientError, Scheme(..))
import qualified Data.ByteString.Lazy as LBS
import qualified Data.List as List
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE
import qualified Data.Text.Lazy as TL

import PUBMED.Types
import PUBMED.ServantClientLogging
import Control.Monad.Reader.Class (asks)

pmHost :: String
pmHost = "eutils.ncbi.nlm.nih.gov"
pmSearchPath :: String
pmSearchPath = "entrez/eutils"
pmPort :: Int
pmPort = 443

defaultClientEnv :: Env ClientEnv
defaultClientEnv = do
  manager' <- liftIO $ newManager tlsManagerSettings
  let env = mkClientEnv manager' $ BaseUrl Https pmHost pmPort pmSearchPath
  enableDebugLogs <- asks enableDebugLogs
  pure $ if enableDebugLogs then addLoggingToClientEnv env else env

-- | API main function
getMetadataWith :: Env (Either Text [PubMed])
getMetadataWith = runSimpleFindPubmedAbstractRequest Nothing

getMetadataWithC :: Env (Either ClientError (Maybe Integer, ConduitT () PubMed IO ()))
getMetadataWithC = do
  config@(Config { apiKey, query, perPage, mWebEnv }) <- ask
  env <- defaultClientEnv
  liftIO $ do
    -- First, estimate the total number of documents
    eRes <- runClientM (searchWithHistory apiKey (Just query) Nothing (Just 1) (Just "y") mWebEnv) env
    pure $ get' config env (fromMaybe defaultPerPage perPage) <$> eRes
  where
    get' :: Config
         -> ClientEnv
         -> PerPage
         -> BsXml
         -> (Maybe Integer, ConduitT () PubMed IO ())
    get' config env perPage (BsXml res) =
      ( Just numResults
      , yieldMany [0..]
        .| takeC (fromInteger numPages)
        .| concatMapMC (\pageNum -> runReaderT (getPage pageNum) $ config { mWebEnv = maybe mWebEnv' Just (mWebEnv  config)}))
      where
        resEncoded = TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict res
        mWebEnv' = parseWebEnv resEncoded
        numResults = fromMaybe 0 $ parseDocCount resEncoded
        numPages = numResults `div` (fromIntegral perPage) + 1

getPage :: Offset
        -> Env [PubMed]
getPage pageNum = do
  (Config { apiKey, perPage = mPerPage, query }) <- ask
  let perPage = fromMaybe defaultPerPage mPerPage
  let offset = fromIntegral $ pageNum * perPage
  debugLog $ "[getPage] getting page " <> show pageNum <> ", offset: " <> show offset <> ", perPage: " <> show perPage <> ", query: " <> T.unpack query <> ", apiKey: " <> show apiKey
  eDocs <- runSimpleFindPubmedAbstractRequest (Just offset)
  case eDocs of
    Left err -> error $ "[getPage] error: " <> show err
    Right docs -> do
      _ <- liftIO $ threadDelay 2_000_000 -- two seconds
      debugLog $ "[getPage] docs length: " <> show (length docs)
      pure docs

runMultipleFPAR :: [Integer]
                -> Env (Either Text [PubMed])
runMultipleFPAR [] = pure $ Right []
runMultipleFPAR ids = do
  (Config { perPage = mPerPage }) <- ask
  let perPage = fromInteger $ fromMaybe defaultPerPage mPerPage
  List.foldl1' concat' <$> mapM runSimpleFetchPubmedAbstractRequest (by perPage ids)
  where
    by _ [] = []
    by n ns = head' : ( by n tail')
      where
        (head', tail') = List.splitAt n ns

    concat' (Right n) (Right m) = Right $ n <> m
    concat' n m = n <> m

runSimpleFetchPubmedAbstractRequest :: [Integer]
                                    -> Env (Either Text [PubMed])
runSimpleFetchPubmedAbstractRequest ids = do
  (Config { apiKey }) <- ask
  env <- defaultClientEnv
  liftIO $ do
    res <- runClientM
      (fetch apiKey (Just "pubmed") (Just "abstract") ids)
      env
    case res of
      (Left err) -> pure (Left . T.pack $ show err)
      (Right (BsXml abs)) -> do
        --putStrLn $ show abs
        case parseOnly removeSub $ LBS.toStrict abs of
          (Left err'') -> pure (Left $ T.pack err'')
          (Right v) -> do
            let parsed = parsePubMed $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict v
            pure $ Right parsed
--          parsed <- catch (pubMedParser v) ((\e -> do
--                                                _ <- debugLog e
--                                                pure []) :: XmlException -> IO [PubMed])
--          Right <$> pure parsed

runSimpleFindPubmedAbstractRequest :: Maybe Offset
                                   -> Env (Either Text [PubMed])
runSimpleFindPubmedAbstractRequest offset =  do
  eDocIds <- searchDocIds offset
  case eDocIds of
    Left err -> pure $ Left err
    Right docIds -> do
      debugLog $ "[runSimpleFindPubmedAbstractRequest] docIds" <> show docIds
      runMultipleFPAR docIds

searchDocIds :: Maybe Offset -> Env (Either Text [Integer])
searchDocIds offset = do
  (Config { apiKey, query, perPage, mWebEnv }) <- ask
  env <- defaultClientEnv
  res <- liftIO $ do
    runClientM
      (searchWithHistory apiKey (Just query) offset perPage (Just "y") mWebEnv)
      env
  case res of
    (Left err) -> pure (Left $ T.pack $ show err)
    (Right (BsXml docs)) -> do
      debugLog $ "[searchDocIds] docs" <> show docs
      pure $ Right $ parseDocIds $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict docs
      --pure $ Right $ runParser parseDocId docs

--runParser :: Show res => (Cursor -> res) -> LBS.ByteString -> res
--runParser parser = parser . fromDocument . parseLBS_ def


-- | TODO this parser need at least one subs at the end
-- (use endOfInput)
removeSub :: Parser ByteString
removeSub = do
  dt <- many textWithBalise
  eo <- manyTill anyChar endOfInput
  pure $ LBS.fromStrict $ pack $ concat dt <> eo
  where
    textWithBalise =
      manyTill anyChar (sub <|> asub)
    sub  = string "<sub>"
      <|> string "<sup>"
      <|> string "<i>"
      <|> string "<b>"
    asub = string "</sub>"
      <|> string "</sup>"
      <|> string "</i>"
      <|> string "</b>"
